/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/sysproc.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/smp_lock.h"
#include "../h/buf.h"
#include "../h/mbuf.h"
#include "../h/un.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/stat.h"
#include "../h/ioctl.h"
#include "../h/uio.h"
#include "../net/net/route.h"
#include "../net/netinet/in.h"
#include "../net/net/if.h"
#include "../h/kmalloc.h"

#include "splice.h"

#define	AVOID_FS	1

#ifdef	SPLICE_DEBUG
#endif
extern u_int	printstate;

/* 
 * splice syscall
 */
splice()
{
	register struct a {
		int	inputfd;
		int	outputfd;
		int	count;
	} *uap = (struct a *)u.u_ap;

	register struct file *fpin, *fpout;

	GETF(fpin, uap->inputfd);
	GETF(fpout, uap->outputfd);

	/*
	 * check for read on the incoming fd
	 * and write on the outgoing fd
	 */

	if ((fpin->f_flag & FREAD) == 0) {
		u.u_error = EBADF;
		return;
	}

	if ((fpout->f_flag & FWRITE) == 0) {
		u.u_error = EBADF;
		return;
	}

#ifdef SPLICE_DEBUG
printstate |= PANICPRINT;
uprintf("splice_syscall: BEGIN... calling splice_begin(0x%x, 0x%x)\n",
fpin, fpout);
#endif


	(void)splice_begin(fpin, fpout); /* will set u.u_error */

#ifdef	SPLICE_DEBUG
uprintf("splice_syscall: RETURNING, errno = %d\n", u.u_error);
printstate &= ~(PANICPRINT);
#endif

	return;
}

int fstrace;
int fstrace_size;
u_char *fstrace_buf;
u_char *fstrace_cur;
struct proc *fstrace_callingproc;

fstrace_ctrl()
{

	register struct a {
		int	sz;
	} *uap = (struct a *)u.u_ap;

	int s = splimp();

	if (uap->sz <= 0) {
		if (fstrace) {
			KM_FREE(fstrace_buf, KM_TEMP);
			fstrace_cur = (u_char *) 0;
		}
		fstrace = 0;
		fstrace_size = 0;
	} else if (uap->sz >= 1) {
		/* if already being traced, doesn't make sense */
		if (fstrace) {
			u.u_error = EINVAL;
			splx(s);
			return;
		}
		KM_ALLOC(fstrace_buf, u_char *, uap->sz,
			KM_TEMP, KM_CLEAR);
		if (!fstrace_buf) {
			u.u_error = ENOMEM;
			splx(s);
			return;
		}
		fstrace = 1;
		fstrace_cur = fstrace_buf;
		fstrace_size = uap->sz;
		fstrace_callingproc = u.u_procp;
	}
	splx(s);
	return;
}

fstrace_fetch()
{

	register struct a {
		u_char	*tracebuf;
		u_int	count;
	} *uap = (struct a *)u.u_ap;
	int s = splimp();
	int sz = MIN(fstrace_size, uap->count);
	if ((fstrace_cur - fstrace_buf) < sz)
		sz = fstrace_cur - fstrace_buf;

	u.u_error = copyout(fstrace_buf, uap->tracebuf, sz);
	splx(s);

	if (!u.u_error)
		u.u_r.r_val1 = sz;
	return;
}

extern struct timeval *timepick;

/* if you change this, be sure to change the user-level code also */
fstrace_rec(str, bp)
	char *str;
	struct buf *bp;
{
	int s;

#ifdef AVOID_FS
return;
#endif
	if (!fstrace)
		return;
	if ((fstrace_cur + sizeof(dev_t) + sizeof(struct timeval))
		>= (fstrace_buf + fstrace_size)) {
cprintf("Warning: fstrace_rec out of bounds\n");
		return;
	}
	/* put a string in first */
	strncpy((char *) fstrace_cur, str, 3);
	fstrace_cur += 4;
	/* then the dev */
	*((dev_t *) fstrace_cur) = bp->b_dev;
	fstrace_cur += sizeof(long);
	*((daddr_t *) fstrace_cur) = bp->b_blkno;
	fstrace_cur += sizeof(long);
	*((long *) fstrace_cur) = bp->b_flags;
	fstrace_cur += sizeof(long);
	*((u_long *) fstrace_cur) = bp->b_lblkno;
	fstrace_cur += sizeof(u_long);
	s = splclock();
	microtime((struct timeval *) fstrace_cur);
#ifdef notdef
	*((struct timeval *) fstrace_cur) = *timepick;
#endif
	splx(s);
	fstrace_cur += sizeof(struct timeval);
	return;
}

/* if you change this, be sure to change the user-level code also */
fstrace_vid(startline)
	int startline;
{
	int s;

#ifdef AVOID_FS
return;
#endif
	if (!fstrace)
		return;
	if ((fstrace_cur + sizeof(dev_t) + sizeof(struct timeval))
		>= (fstrace_buf + fstrace_size)) {
cprintf("Warning: fstrace_rec out of bounds\n");
		return;
	}
	*((int *) fstrace_cur) = startline;
	fstrace_cur += sizeof(int);
	s = splclock();
	microtime((struct timeval *) fstrace_cur);
#ifdef notdef
	*((struct timeval *) fstrace_cur) = *timepick;
#endif
	splx(s);
	fstrace_cur += sizeof(struct timeval);
	return;
}

/* called from lower half, so don't do splx */
ntrace_rec(str, m)
	char *str; /* at most 3 chars */
	struct mbuf *m;
{
	if (!fstrace)
		return;
	if ((fstrace_cur + sizeof(struct timeval) + 4 + sizeof(int))
		>= (fstrace_buf + fstrace_size)) {
cprintf("Warning: fstrace_rec out of bounds\n");
		return;
	}
	/* write time record first */
	microtime((struct timeval *) fstrace_cur);
	fstrace_cur += sizeof(struct timeval);

	/* put a string in next */
	strncpy((char *) fstrace_cur, str, 3);
	fstrace_cur += 4;

	/* put an int next */
	if (m)
		*((int *) fstrace_cur) = m->m_len;
	else
		*((int *) fstrace_cur) = 0;
	fstrace_cur += sizeof(int);

	return;
}

ntrace_tophalf_rec(str, code)
	char *str;
	int code;

{
	int s;
#ifdef notdef
	if (!fstrace || (u.u_procp != fstrace_callingproc))
		return;
#endif
	if (!fstrace)
		return;

	s = splimp();
	if ((fstrace_cur + sizeof(struct timeval) + 4 + sizeof(int) )
		>= (fstrace_buf + fstrace_size)) {
cprintf("Warning: fstrace_rec out of bounds\n");
		splx(s);
		return;
	}
	/* write time record first */
	microtime((struct timeval *) fstrace_cur);
	fstrace_cur += sizeof(struct timeval);

	/* put a string in next */
	strncpy((char *) fstrace_cur, str, 3);
	fstrace_cur += 4;

	/* put an int next */
	*((int *) fstrace_cur) = code;
	fstrace_cur += sizeof(int);

	splx(s);
	return;
}
