/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
 * The splice descriptor...
 *	just about everything needed to fully identify a splice
 *	note: is pointed to by a new field in the buf structure
 *	called b_splicep
 */

struct splice_descriptor {

	struct splice_descriptor *sp_next;	/* link on chain */
	struct splice_descriptor *sp_prev;	/* forw link */
	u_int sp_state;				/* see below */
	struct proc *sp_proc;			/* caller */
	struct exit_actn *sp_xa;		/* action during exit */
	int sp_owrites;				/* # pending writes */
	int sp_oreads;				/* # pending reads */

	/*
	 * Note: apparently the compiler WON'T pad,
	 * so be careful to put shorts at the end! 
	 */

	/*
	 * SRC DESCRIPTORS
	 */

	struct file *sp_srcfile;
	union {
		struct gnode *sp_Srcgp;   /* src file gnode ptr */
		struct socket *sp_Srcsock; /* src socket ptr */
	} sp_SRC;
	int sp_srcoffset;	/* src offset (bytes) */

	/*
	 * DEST DESCRIPTORS
	 */

	struct file *sp_destfile;
	union {
		struct gnode *sp_Destgp; /* dest file gnode ptr */
		struct socket *sp_Destsock; /* dest socket */
	} sp_DEST;

	/*
	 * MISC INFO ABOUT SRC or DEST
	 */

	u_long sp_destlastlbn;
	union {
		daddr_t *sp_Srcblockmap; /* list of phy block #s on src */
		int (*sp_Srccdevget)();
	} sp_SRCINFO;
	union {
		daddr_t *sp_Destblockmap; /* list of phy block #s on dest */
		int (*sp_Destdevput)();
	} sp_DESTINFO;
};

/* structure used in csplicesw, similar to cdevsw */
struct splice_cdev {
	int (*getfn)();
	int (*putfn)();
};

/* defines for "data hiding" (to dig into unions) */
#define	sp_srcgp	sp_SRC.sp_Srcgp
#define	sp_srcsock	sp_SRC.sp_Srcsock
#define	sp_destgp	sp_DEST.sp_Destgp
#define	sp_destsock	sp_DEST.sp_Destsock

#define	sp_srcblockmap	sp_SRCINFO.sp_Srcblockmap
#define	sp_destblockmap	sp_DESTINFO.sp_Destblockmap

#define sp_cdevget	sp_SRCINFO.sp_Srccdevget
#define sp_cdevput	sp_DESTINFO.sp_Destdevput

#define	ISSOCK(fp)	(fp->f_type == DTYPE_SOCKET)
#define	ISINODE(fp)	(fp->f_type == DTYPE_INODE)
#define	ISREG(gp)	((gp->g_mode & GFMT) == GFREG)
#define	ISCHR(gp)	((gp->g_mode & GFMT) == GFCHR)

/* all the splice routines */
extern struct buf *splice_bread();
extern struct buf *splice_breada();
extern struct buf *splice_newbuf();
extern struct buf *splice_getblk();
extern int splice_begin();
extern int splice_read();
extern int splice_write();
extern int splice_read_handler();
extern int splice_write_handler();
extern int splice_exit_handler();
extern int splice_establish_procexit();
extern int splice_free();
extern int splice_rw();

extern struct splice_descriptor *find_splice_xa();
extern struct splice_descriptor *find_splice_proc();
extern struct splice_descriptor *splice_alloc();

extern struct splice_cdev csplicesw[];

/* return values from splice_read() */

#define	SPLICE_READ_ERROR	1	/* error */
#define	SPLICE_READ_HIT		2	/* direct blk, cache hit */
#define	SPLICE_READ_MISS	3	/* direct blk, cache miss */ 
#define	SPLICE_READ_EOF		4	/* eof on input */

/* return values from splice_write() */ 
#define	SPLICE_WRITE_ERROR	1	/* write err */
#define	SPLICE_WRITE_OK		2	/* async write scheduled */
#define	SPLICE_WRITE_MISS	3	/* problem w/getblk? */

/* state bits for sp->sp_state */

#define	SPLICE_ACTIVE		0x00000001	/* data being moved */
#define	SPLICE_WANTED		0x00000002	/* proc waiting */
#define	SPLICE_EOF		0x00000004	/* eof on input */
#define	SPLICE_NO_OUTCKSUM	0x00000008	/* no cksum on outpt */

/* other constants */

#define	SPLICE_MAXOREADS	5		/* max# pending reads */
#define	SPLICE_MINREADS		3		/* min# pending reads */
#define	SPLICE_MAXOWRITES	5		/* max# pending wr */
#define	SPLICE_NCDEVS		100		/* #chrdev entries */

