/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/sysproc.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/smp_lock.h"
#include "../h/buf.h"
#include "../h/mbuf.h"
#include "../h/un.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/stat.h"
#include "../h/ioctl.h"
#include "../h/uio.h"
#include "../net/net/route.h"
#include "../net/netinet/in.h"
#include "../net/net/if.h"
#include "../h/kmalloc.h"

#include "../csl/splice.h"

/*
 * splice_rw()
 *
 *	perform read/write loop until a cache miss on the read
 *
 *	Note: this assumes splice_read (actually splice_bread)
 *	will set up the callback to splice_read_handler for us, so
 *	we don't have to do it here
 */

int
splice_rw(sp)
	struct splice_descriptor *sp;
{
	int rval, wval;
	int s;
	struct buf *bp;

	while (1) {

		rval = splice_read(sp, &bp);

#ifdef SPLICE_DEBUG
cprintf("splice_rw: splice_read returned %d\n", rval);
#endif
		switch (rval) {

		case SPLICE_READ_ERROR:
			panic("splice_rw: read error");
			return (-1); /* NOTREACHED */

		case SPLICE_READ_HIT:
			s = splbio();
			sp->sp_owrites++;
			splx(s);
			wval = splice_write(bp);

			switch(wval) {
			case SPLICE_WRITE_ERROR:
				panic("splice_rw: writeerr");
			case SPLICE_WRITE_MISS:
				panic("splice_rw: writemiss");
			case SPLICE_WRITE_OK:
				continue;
			default:
				panic("unknown retval s_write");
			}

		case SPLICE_READ_MISS:
			return (0);

		case SPLICE_READ_EOF:

			sp->sp_state |= SPLICE_EOF;

			/*
			 * if there's an outstanding write/read, the
			 * write/read handler will take care of it
			 */
			 
			if (sp->sp_owrites || sp->sp_oreads) {
				return(1);
			}

			if (sp->sp_state&SPLICE_EOF) {
				sp->sp_state &= ~(SPLICE_ACTIVE);
				if ((sp->sp_state&SPLICE_WANTED))
					wakeup((caddr_t) sp);
			}
			return (1);

		default:
			panic("splice_rw: rval from splice_read");
		}
	}
	/* NOTREACHED */
}

/*
 * setup_callback()
 * 	install an "I/O complete" handler for a given buffer
 *
 *	if this is for a splice decriptor, also fill in the
 *	field in the buf header with the splice pointer
 *
 *	The handler function will be invoked by biodone()
 */

void
setup_callback(bp, sp, handler)
	struct buf *bp;
	struct splice_descriptor *sp;
	int (*handler)();
{

#ifdef	SPLICE_DEBUG
cprintf("setup_callback: handler@0x%x, buf@0x%x\n", handler, bp);
#endif

	bp->b_flags |= B_CALL;
	bp->b_iodone = handler;
	/* 
	 * didn't want to have to change buf.h to really have
	 * a pointer to a struct splice, as I'd have to
	 * change every source file in the system to include splice.h,
	 * which I wanted to avoid... SO, instead use u_char * in buf.h,
	 * and assume it's interchangeable with a (struct splice_desc *)
	 */
	if (sp != (struct splice_descriptor *) NULL)
		bp->b_splicep = (u_char *) sp;
}

#if (defined(TRACE) || defined(SYS_TRACE)) && !defined(trace1)
trace1(id, arg1, arg2)
	int id, arg1, arg2;
{
	cprintf("TRACEPOINT: trace_id=%d, arg1=%d, arg2=%d\n",
		id, arg1, arg2);
}
#endif
