
/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "../machine/pte.h"
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/proc.h"
#include "../h/seg.h"
#include "../h/smp_lock.h"
#include "../h/vm.h"
#include "../h/trace.h"
#include "../h/gnode.h"
#include "../h/mount.h"
#include "../h/cpudata.h"

#include "../h/mbuf.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/errno.h"
#include "../h/time.h"

#include "../net/net/if.h"
#include "../net/net/route.h"

#include "../net/netinet/in.h"
#include "../net/netinet/in_pcb.h"
#include "../net/netinet/in_systm.h"
#include "../net/netinet/in_var.h"
#include "../net/netinet/ip.h"
#include "../net/netinet/ip_var.h"

#include "splice.h"

extern struct bufstats bufstats;

/*
 * splice_newbuf
 *	very much like getnewbuf, except looks on EMPTY list 1st
 *
 * Find a buffer header which is available for use.
 * Select something from a free list.
 *
 *	search the EMPTY list (no mapped memory)
 */
struct buf *
splice_newbuf(sp)
	struct splice_descriptor *sp;
{
	register struct buf *bp, *dp;
	register int s;

loop:
	s = splbio();
	smp_lock(&lk_bio, LK_RETRY);
	/*
	 * this next loop checks all queues backwards
	 * starting with BQ_EMPTY(3) and going down to
	 * BW_DIRTY(1)
	 */
	for (dp = &bfreelist[BQ_EMPTY]; dp > bfreelist; dp--)
		if (dp->av_forw != dp)
			break;

#ifdef SPLICE_DEBUG
cprintf("splice_newbuf: got buffer from ");
if (dp == &bfreelist[BQ_EMPTY])
	cprintf(" empty list\n");
else if (dp == &bfreelist[BQ_CLEAN])
	cprintf(" clean list\n");
else if (dp == &bfreelist[BQ_DIRTY])
	cprintf(" dirty list\n");
else
	cprintf(" unknown list 0x%x\n", dp);
#endif

	if (dp == bfreelist) {		/* NO free blocks on any */
		int i;
		splx(s);
cprintf("sp_state=0x%x, owrites=%d, oreads=%d, srcoff=%d\n",
sp->sp_state, sp->sp_owrites, sp->sp_oreads, sp->sp_srcoffset);
		for (i = 0; i < 100000000; i++)
			;
		panic("splice_newbuf: no buffer available free lists");
	}

	/*
	 * ok, found something
	 */

	bp = dp->av_forw;
	bp->state = 0;

	/*
	 * Remove the buffer from the free list and push it out if delayed.
	 */
	notavail(bp);
	smp_unlock(&lk_bio);
	splx(s);
	if (bp->b_flags & B_DELWRI) {
		bp->b_flags |= B_ASYNC;
		bp->b_flags |= B_FORCEWRITE;
		++bufstats.delwrite;
		++bufstats.forcewrite;
cprintf("splice_newbuf: found del-write buf@0x%x", bp); 
		bawrite(bp); /* changed from bwrite */
		goto loop;
	}
	trace(TR_BRELSE, bp->b_dev, bp->b_blkno);
	++bufstats.newbuf;

	/*
	 * Clean out the flag word, but leave the B_BUSY bit 
	 * (notavail set state)
	 */
	bp->b_flags = B_BUSY;

	return (bp);
}



/*
 * Assign a buffer for the given block.  If the appropriate
 * block is already associated, return it; otherwise search
 * for the oldest non-busy buffer and reassign it.
 *
 * For splice: we don't really care if there's any memory
 * associated with the buf header, or how big it is
 */
extern int getblk_mp_window;

struct buf *
splice_getblk(dev, blkno, size, gp, sp)
	dev_t dev;
	register daddr_t blkno;
	register int size;
	register struct gnode *gp;
	struct splice_descriptor *sp;
{
	register struct buf *bp, *dp;
	register int s;

#ifdef SPLICE_DEBUG
cprintf("splice_getblk:dev=%d, blkno=%d, size=%d, gp=0x%x\n",
(int)dev, blkno, size, gp);
#endif

ULTRIX CODE DELETED
#ifdef SPLICE
			bp->b_bcount = size;
#else
			if (brealloc(bp, size) == 0)
				goto loop;
#endif

	/*
	 * when we've reached this point, we *didn't* find the buffer
	 * in the cache, so get a new header
	 * splice_newbuf will try the empty list first,
	 * and will never return us a delayed-write buffer
	 */

	bp = splice_newbuf(sp);	/* new header from EMPTY list */

#ifdef SPLICE
	bp->b_bcount = size;
#else
	if (brealloc(bp, size) == 0)
		goto loop;
#endif

#ifdef SPLICE_DEBUG
cprintf("splice_getblk, returning NEW buf@0x%x\n", bp);
#endif
	return(bp);
}
