
/*
 * Copyright (c) 1992, 1993, 1994, 1995
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, San Diego and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/sysproc.h"
#include "../h/file.h"
#include "../h/inode.h"
#include "../h/smp_lock.h"
#include "../h/buf.h"
#include "../h/mbuf.h"
#include "../h/un.h"
#include "../h/conf.h"
#include "../h/mount.h"
#include "../h/gnode_common.h"
#include "../h/gnode.h"
#include "../h/buf.h"
#include "../h/fs.h"
#include "../fs/ufs/fs.h"
#include "../fs/ufs/ufs_inode.h"
#include "../h/domain.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/stat.h"
#include "../h/ioctl.h"
#include "../h/uio.h"
#include "../net/net/route.h"
#include "../net/netinet/in.h"
#include "../net/net/if.h"
#include "../h/kmalloc.h"
#include "../h/cmap.h"
#include "../h/ksched.h"
#include "splice.h"


/*
 * splice_write:
 *	basically emulates the write part of rwip. (ufs_rwgp in ultrix)
 *	challenge: must allocate a buf header and tack
 *	on the data
 */

int splice_write_handler();	/* called when write completes */

int
splice_write(inbp)
	struct buf *inbp;		/* buf ptr from read() */
{
	struct splice_descriptor *sp;	/* connection point */
	struct gnode *gp;
	dev_t dev;
	struct fs *fs;
	long bsize;
	daddr_t lbn; /* assumes *same* lblock size on both fs's */
	daddr_t pbn;
	int size;

	int n, s;
	struct buf *bp;		/* new bufheader from EMPTY list */

	if ((!inbp) || (!inbp->b_splicep))
		panic("splice_write: no blk to write");

	sp = (struct splice_descriptor *) inbp->b_splicep;
	gp = sp->sp_destgp;
	dev = (dev_t) gp->g_dev;
	fs = FS(gp);
	bsize = fs->fs_bsize;
	lbn = inbp->b_lblkno;
	size = blksize(fs, gp, lbn);

	if (lbn > sp->sp_destlastlbn)
		panic("splice_write: lbn out of range");
	pbn = sp->sp_destblockmap[(int) lbn];


	/*
	 * we may be undersized if we're the last block
	 */

	if (lbn != sp->sp_destlastlbn)
		n = bsize;
	else
		n = sp->sp_srcgp->g_size % bsize;

#ifdef SPLICE_DEBUG
cprintf("splice_write:lbn=%d,pbn=%d,n=%d,size=%d,bsize=%ld,g_size=%d\n",
lbn, (int)pbn, n, size, bsize, (int)gp->g_size);
#endif

	if ((int)pbn <= 0)
		panic("splice_write: pbn out of range");

	if (n <= 0)
		panic("splice_write: n size for blk <= 0");

	/*
	 * get a buf header for the destination
	 *	normal case with be the first one; we only
	 *	would ever do the second if we're re-writing
	 *	the middle of a block, which the current version
	 *	of splice isn't capable of (kfall, 6/92)
	 */

	bp = splice_getblk(dev, pbn, size, (struct gnode *) NULL, sp);
	if (bp == NULL)
		return (SPLICE_WRITE_MISS);	/* ends up as panic */

#ifdef notdef
if (n == bsize)
	bp = splice_getblk(dev, pbn, size, (struct gnode *) NULL);
else
	bp = bread(dev, pbn, size, (struct gnode *) NULL);
#endif

	/*
	 *	splice_getblk will have returned a ready
	 *	buffer directly, or fetched a new one.
	 *	If it's a new one, it will have first
	 *	looked on the empty list and then on
	 *	the others.
	 *
	 *	If we ended up with a buffer from the empty
	 *	list, meaning its memory won't be mapped, but
	 *	the vaddr (in bp->b_un.b_addr) is still ok,
	 *	so reassign it. 
	 */

	bp->b_inbp = inbp; /* so write handler can brelse(inbp) */

	/* save copy of info we just got */
	bp->b_savep = bp->b_un.b_addr; /* copy of vaddr from getblk */
	bp->b_savesize = bp->b_bufsize;	/* save size */
	bp->b_savebcount = bp->b_bcount; /* i/o cnt size */
	bp->b_saveflags = bp->b_saveflags;

	/* make this buf header point to the source buf hdr */
	bp->b_un.b_addr = inbp->b_un.b_addr;
	bp->b_bufsize = inbp->b_bufsize;
	bp->b_bcount = inbp->b_bcount;
/* bp->b_flags |= B_CACHE;	it's cached */
	bp->b_lblkno = inbp->b_lblkno;

/* XXsomewhere here should deal with sub-block length write */
/* to do this, must bread block first, make deltas, and rewrite */


	setup_callback(bp, sp, splice_write_handler);

#ifdef SPLICE_DEBUG
cprintf("splice_write: calling bawrite(buf@0x%x)..\n", bp);
#endif

	/* bawrite() is a macro in buf.h */
	bawrite(bp);	/* async but not delayed write */
			/* sets ASYNC flag, and calls bwrite() */

	return (SPLICE_WRITE_OK);
}

/*
 * splice_write_handler:
 *	this code is invoked when a bawrite completes,
 *	it is called via biodone()
 *
 *	biodone() will have set the B_DONE field in b_flags for
 *	us
 *
 *	biodone() will also have turned off the B_CALL field
 *	in b_flags
 */

int
splice_write_handler(bp)
	struct buf *bp;
{
	register i;

	struct splice_descriptor *sp = (struct splice_descriptor *)
			bp->b_splicep;
	struct buf *inbp = bp->b_inbp;

	if (bp->b_flags & B_ERROR)
		panic("splice_write_hadler: IO error in bawrite()");

	if (bp->state & B_WANTED)
		panic("splice_write_handler: buf was WANTED");


#ifdef SPLICE_DEBUG
cprintf("splice_write_handler: buf@0x%x, inbp@0x%x, lbn=%d, nwrites=%d, nreads=%d\n",
bp, bp->b_inbp, bp->b_lblkno, sp->sp_owrites, sp->sp_oreads);
#endif

	/*
	 * we assume here a write has completed successfully
	 * idea is to restore it back to what it looked like when
	 * it was allocated (from EMPTY list)
	 */

	bp->b_un.b_addr = (caddr_t) bp->b_savep;
	bp->b_bufsize = bp->b_savesize;
	bp->b_flags = bp->b_saveflags;

	/*
	 * logic in brelse() dictates that if b_bufsize <=0,
	 * then the block will be inserted at the head of the
	 * EMPTY list
	 *
	 * we must turn on the INVAL bit, because now that we've
	 * restored the b_addr field above, we are pointing to the
	 * original buffer, and it never was involved in I/O
	 */

	bp->state |= B_INVAL;
	bp->b_lblkno = 0;
	brelse(bp);	/* return new buffer to empty list */

	 /*
	  * we've completed the write, now free up the
	  * incoming (read) side
	  *
	  * Comment: if we turn on the INVAL or NOCACHE bit
	  * here before the brelse we can probably avoid the
	  * data just read filling up the cache!
	  */

	inbp->b_lblkno = 0L;
	brelse(inbp);

#ifdef notdef
Does this belong here?
bremhash(inbp);
#endif

	--sp->sp_owrites;	/* this buf completed */

	/* if we are too low, give us some more */
	if (sp->sp_oreads < SPLICE_MINREADS && 
	    (sp->sp_owrites < SPLICE_MAXOWRITES)) {

		for (i = SPLICE_MAXOREADS; i; i--) {
			if (splice_rw(sp) == 1)
				break;
		}

		{
			struct kschedblk kblk;
			/* schedule subsequent reads */
			kblk.arg = (caddr_t) sp;
			kblk.func = (void (*)()) splice_rw;
			ksched(&kblk);
		}

	} /* if (we're low on read side) */

	if (sp->sp_owrites < 0)	/* shouldn't happen */
		panic("splice_write_handler: owrites < 0");

	/*
	 *
	 * we are completely finished if
	 *	1> EOF is reached on the input AND
	 *	2> there's no more outstanding writes
	 *
	 * if we are completely finished
	 *	1> turn off active state
	 *	2> if there's somebody waiting, wake it up
	 *	3> otherwise, call splice_finish() ourselves
	 */

	 if ((sp->sp_owrites == 0) && (sp->sp_oreads == 0) &&
	    (sp->sp_state&SPLICE_EOF)) {
		sp->sp_state &= ~(SPLICE_ACTIVE);
		if (sp->sp_state&SPLICE_WANTED)
			wakeup((caddr_t)sp);
	}

	return(0); /* next read will happen by splice_rw() */
}


/*
 * debug routine to dump a buffer header and the 1st 3 bytes
 * of its data contents
 */

#define	BYTE(x)	((x)&0xff)

buf_debug(bp)
	struct buf *bp;
{
	u_char *p = (u_char *) bp->b_un.b_addr;

	uprintf("===== buf@0x%x =====\n", bp);
	uprintf("b_flags = 0x%x,", bp->b_flags);
	uprintf(" b_bcount = %d,", bp->b_bcount);
	uprintf(" b_bufsize = %d\n", bp->b_bufsize);
	uprintf("b_error = %d, ", (int)bp->b_error);
	uprintf("b_dev = %d,", (int)bp->b_dev);
	uprintf("   major: %d, minor: %d\n",
			(int)major(bp->b_dev), (int)minor(bp->b_dev));
	uprintf("b_un.b_addr = 0x%x, ", p);
	uprintf("	Bytes: '%c' (0x%x) '%c' (0x%x) '%c' (0x%x)\n",
		BYTE(*p), BYTE(*p),
		BYTE(p[1]), BYTE(p[1]),
		BYTE(p[2]), BYTE(p[2]));
	uprintf("b_blkno = %d, ", bp->b_blkno);
	uprintf("b_resid = %ld, ", bp->b_resid);
	uprintf("b_iodone[handler] -> 0x%x, ", bp->b_iodone);
	uprintf("state = 0x%x\n", bp->state);
	uprintf("b_savep = 0x%x, ", bp->b_savep);
	uprintf("b_savesize = %ld\n", bp->b_savesize);
	uprintf("===================\n");
}

#ifdef	SPLICE_DEBUG

lock_debug()
{
	register i;
	for (i = lowcpu; i <= highcpu; i++) {
		if (CPUDATA(i) == 0)
			continue;
		print_locks_held(i);
		print_lock_trace(i);
	}
	print_process_locks();
}
#endif
